<?php

/**
 * @file
 * Contains \Drupal\couchbasedrupal\Cache\CouchbaseBackend.
 */

namespace Drupal\couchbasedrupal\Cache;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Cache\CacheTagsChecksumInterface;
use Drupal\Core\Cache\Cache;

/**
 * Stores cache items in Couchbase.
 */
class CouchbaseBackend implements CacheBackendInterface {

  /**
   * The name of the cache bin to use.
   *
   * @var string
   */
  protected $bin;

  /**
   * Prefix for the site.
   *
   * @var string
   */
  protected $sitePrefix;


  /**
   * Wether to do reliable reads. "Slower" but consistent.
   *
   * @var boolean
   */
  protected $reliable = TRUE;

  /**
   * Prefix for all keys in this cache bin.
   *
   * Includes the site-specific prefix in $sitePrefix.
   *
   * @var string
   */
  protected $binPrefix;

  /**
   * Couchbase bucket
   *
   * @var \CouchbaseBucket
   */
  protected $bucket;


  /**
   * Query options.
   *
   * @var array
   */
  protected $options = array();

  /**
   * Name of the view for this binary in the Couchbase database
   *
   * @var string
   */
  protected $view;

  /**
   * The cache tags checksum provider.
   *
   * @var \Drupal\Core\Cache\CacheTagsChecksumInterface
   */
  protected $checksumProvider;

  /**
   * Last time this cache was cleared.
   *
   * @var float
   */
  protected $lastClearTime = 0;

  /**
   * Make sure that we create a view for each binary.
   *
   * Not used for anything yet...
   *
   * @throws \Exception
   */
  protected function EnsureViews() {

    // WARNING!! getDesignDocument and upsertDesignDocument are extremely slow operations!!
    //// we need to cache this somewhere....
    if(\Drupal::cache('couchbase')->get('view_exists:' . $this->binPrefix)) {
      return;
    }

    /** @var \CouchbaseBucketManager */
    $manager = $this->bucket->manager();

    $dd = $manager->getDesignDocument($this->binPrefix);

    if (empty($dd)) {

      $document = array(
          'views' => array(
            // Used to getAll() items in this binary.
            'keys' => array(
              'map' => 'function (doc, meta) { if (doc.binPrefix == "' . $this->binPrefix .'") { emit(meta.id, doc.cid); } }'
            ),
            // Used to garbage collect deleted items.
            'created' => array(
              'map' => 'function (doc, meta) { if (doc.binPrefix == "' . $this->binPrefix .'") { emit(doc.created, doc.cid); } }'
            ),
          )
        );

      $map = json_encode($document);

      $manager->upsertDesignDocument($this->binPrefix, $map);

      // TODO: upsertDesignDocument always returns true... even if the creation fails.
      if ($manager->getDesignDocument($this->binPrefix) == FALSE) {
        throw new \Exception("Could not create view.");
      }
    }

    \Drupal::cache('couchbase')->set('view_exists:' . $this->binPrefix, TRUE);
  }

  /**
   * Get or Set the last clear time.
   *
   * @param float $time
   */
  protected function LastClear($time = NULL) {
    $key = $this->sitePrefix . ':' . $this->bin . '-lastcleared';
    if (empty($time)) {
      if ($cache = $this->bucketGet($key)) {
        $this->lastClearTime = $cache->value;
      }
      $this->lastClearTime = 0;
    }
    else {
      $this->bucketUpsert($key, $time);
      $this->lastClearTime = $time;
    }
  }

  /**
   * Constructs a new CouchbaseBackend instance.
   *
   * @param string $bin
   *   The name of the cache bin.
   * @param string $site_prefix
   *   The prefix to use for all keys in the storage that belong to this site.
   * @param \Drupal\Core\Cache\CacheTagsChecksumInterface $checksum_provider
   *   The cache tags checksum provider.
   */
  public function __construct($bin, $sitePrefix, CacheTagsChecksumInterface $checksum_provider, \CouchbaseBucket $bucket) {
    $this->bin = $bin;
    $this->checksumProvider = $checksum_provider;
    $this->sitePrefix = $sitePrefix;
    // We need to be this as short as possible, but the binary should be human readable
    // when possible.
    $this->binPrefix = crc32($this->sitePrefix) . ':' . $bin . ':';
    $this->bucket = $bucket;
    $this->view = $this->binPrefix;
    $this->options = array();
    $this->LastClear();
    $this->EnsureViews();
  }

  /**
   * Prepends the Wincache user variable prefix for this bin to a cache item ID.
   *
   * @param string $cid
   *   The cache item ID to prefix.
   *
   * @return string
   *   The APCu key for the cache item ID.
   */
  protected function getBinKey($cid) {
    $key = $this->binPrefix . $cid;
    // If you surpas 250 bytes the PHP drivers HAAAAANGS
    if (mb_strlen($key, '8bit') >= 250) {
      $key = hash('sha256', $this->binPrefix . $cid);
    }
    return $key;
  }

  /**
   * {@inheritdoc}
   */
  public function get($cid, $allow_invalid = FALSE) {
    $cids = array($cid);
    $cache = $this->getMultiple($cids, $allow_invalid);
    return reset($cache);
  }

  /**
   * {@inheritdoc}
   */
  public function getMultiple(&$cids, $allow_invalid = FALSE) {
    // Translate the requested cache item IDs to Wincache keys.
    $map = array();
    foreach ($cids as $cid) {
      $map[$this->getBinKey($cid)] = $cid;
    }

    $result = $this->bucket->get(array_keys($map));
    $cache = array();
    foreach ($result as $key => $item) {
      $item = $this->prepareItem($item->value, $allow_invalid);
      if ($item) {
        $cache[$map[$key]] = $item;
      }
    }
    unset($result);

    $cids = array_diff($cids, array_keys($cache));
    return $cache;
  }

  /**
   * Returns all cached items, optionally limited by a cache ID prefix.
   *
   *
   * @param string $prefix
   *   (optional) A cache ID prefix to limit the result to.
   *
   * @return string[]
   *   An array of Keys for this binary.
   */
  public function getAll($prefix = '') {
    $query = \CouchbaseViewQuery::from($this->view, 'keys'); //->custom(array('startkey' => "%22$prefix%22"));
    $rows = $this->bucket->query($query)['rows'];
    return array_column($rows, 'value');
  }

  /**
   * Returns all cached items for this binary.
   *
   * @return string[]
   */
  protected function getAllBinary() {
    throw new \Exception("Not implemented.");
  }

  /**
   * Prepares a cached item.
   *
   * Checks that the item is either permanent or did not expire.
   *
   * @param \stdClass $cache
   *   An item loaded from cache_get() or cache_get_multiple().
   * @param bool $allow_invalid
   *   If TRUE, a cache item may be returned even if it is expired or has been
   *   invalidated. See ::get().
   *
   * @return mixed
   *   The cache item or FALSE if the item expired.
   */
  protected function prepareItem($cache, $allow_invalid) {
    if (!isset($cache->data)) {
      return FALSE;
    }

    if ($cache->created <= $this->lastClearTime && $allow_invalid == FALSE) {
      return FALSE;
    }

    $cache->tags = $cache->tags ? explode(' ', $cache->tags) : array();

    // Check expire time.
    $cache->valid = $cache->expire == Cache::PERMANENT || $cache->expire >= REQUEST_TIME;

    // Check if invalidateTags() has been called with any of the entry's tags.
    if (!$this->checksumProvider->isValid($cache->checksum, $cache->tags)) {
      $cache->valid = FALSE;
    }

    // Unserialize if value was serialized
    if ($cache->serialized == TRUE) {
      $cache->data = @unserialize($cache->data);
      // If we fail to unserialize, this is a cache miss.
      if ($cache->data == FALSE) {
        return FALSE;
      }
    }

    if (!$allow_invalid && !$cache->valid) {
      return FALSE;
    }

    return $cache;
  }

  protected function prepareCacheItem($cid, $data, $expire, array $tags = array()) {
    Cache::validateTags($tags);
    $tags = array_unique($tags);
    $cache = new \stdClass();
    $cache->cid = $cid;
    $cache->binPrefix = $this->binPrefix;
    $cache->created = round(microtime(TRUE), 3);
    $cache->expire = $expire;
    $cache->tags = implode(' ', $tags);
    $cache->checksum = $this->checksumProvider->getCurrentChecksum($tags);
    $cache->serialized = FALSE;
    $cache->data = $data;
    if (!is_scalar($data)) {
      $cache->serialized = TRUE;
      $cache->data = serialize($cache->data);
    }
    return $cache;
  }



  /**
   * {@inheritdoc}
   */
  public function set($cid, $data, $expire = CacheBackendInterface::CACHE_PERMANENT, array $tags = array()) {

    $cache = $this->prepareCacheItem($cid, $data, $expire, $tags);

    // If you don't pass an int couchbase crashes.
    $expire = (int) $expire;

    // Omit the expire parameter for permanent items.
    if ($expire == CacheBackendInterface::CACHE_PERMANENT) {
      $this->bucketUpsert($this->getBinKey($cid), $cache);
    }
    else {
      $this->bucketUpsert($this->getBinKey($cid), $cache, array('expiry' => $expire));
    }
  }

  /**
   * Like set() but will return false (and fail) if the document already exists.
   */
  public function add($cid, $data, $expire = CacheBackendInterface::CACHE_PERMANENT, array $tags = array()) {
    // If you don't pass an INT couchbase crashes.
    $expire = (int) $expire;
    // Prepare the item.
    $cache = $this->prepareCacheItem($cid, $data, $expire, $tags);
    $set_result = TRUE;
    if ($expire == CacheBackendInterface::CACHE_PERMANENT) {
      $set_result = $this->bucketInsert($this->getBinKey($cid), array('expiry' =>  $expire));
    }
    else {
      $set_result = $this->bucketInsert($this->getBinKey($cid), $cache, array('expiry' => $expire));
    }
    return $set_result;
  }

  /**
   * {@inheritdoc}
   */
  public function setMultiple(array $items = array()) {
    foreach ($items as $cid => $item) {
      $this->set($cid, $item['data'], isset($item['expire']) ? $item['expire'] : CacheBackendInterface::CACHE_PERMANENT, isset($item['tags']) ? $item['tags'] : array());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function delete($cid) {
    $this->bucketRemove($this->getBinKey($cid));
  }

  /**
   * {@inheritdoc}
   */
  public function deleteMultiple(array $cids) {
    $keys = array_map(array($this, 'getBinKey'), $cids);
    foreach ($keys as $key) {
      $this->bucketRemove($key);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deleteAll() {
    $this->LastClear(microtime(TRUE));
  }

  /**
   * Retrieve an item from the bucket without throwing
   * an Exception if it does not exist.
   *
   * @param string|string[] $key
   * @return mixed|bool
   */
  protected function bucketGet($key) {
    try {
      return $this->bucket->get($key);
    }
    catch (\Exception $e) {
      if ($e->getCode() == 13) {
        return FALSE;
      }
      throw $e;
    }
  }

  /**
   * Remove an item from the bucket without
   * throwing and exceptino if it does not exist.
   *
   * @param string|string[] $key
   * @return mixed|bool
   */
  protected function bucketRemove($key) {
    try {
      if (!is_array($key)) {
        $key = array($key);
      }
      foreach ($key as $id) {
        return $this->bucket->remove($id);
      }
    }
    catch (\Exception $e) {
      if ($e->getCode() == 13) {
        return FALSE;
      }
      throw $e;
    }
    return TRUE;
  }

  private function bucketUpsert($id, $val = NULL, $options = array()) {
    $this->bucket->upsert($id, $val, $options);
  }

  /**
   * Insert an item in a bucket. On failure or if the item already exists
   * returns FALSE.
   * 
   * @param string $id 
   * @param mixed $val 
   * @param mixed $options 
   * @return bool
   */
  private function bucketInsert($id, $val = NULL, $options = array()) {
    try {
      $this->bucket->insert($id, $val, $options);
      return TRUE;
    }
    catch (\Exception $e) {
      $code = $e->getCode();
      // 12,13 are document already exists or document does not exist.
      if (in_array($code, array(12, 13))) {
        return FALSE;
      }
      throw $e;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function garbageCollection() {
    // This is not working...
    return;
    $query = \CouchbaseViewQuery::from($this->view, 'created');
    $query->range(0, $this->lastClearTime);
    $batch_size = 50;
    while (TRUE) {
      $query->limit($batch_size);
      $results = $this->bucket->query($query, array('stale' => FALSE));
      $total = count($results['rows']);
      if ($total == 0) {
        break;
      }
      foreach($results['rows'] as $row) {
        $this->delete($row['value']);
      }
      if ($total < $batch_size) {
        break;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function removeBin() {
    $this->bucketRemove($this->getAllBinary());
  }

  /**
   * {@inheritdoc}
   */
  public function invalidate($cid) {
    $this->invalidateMultiple(array($cid));
  }

  /**
   * {@inheritdoc}
   */
  public function invalidateMultiple(array $cids) {
    foreach ($this->getMultiple($cids) as $cache) {
      $this->bucket->touch($this->getBinKey($cache->cid), REQUEST_TIME - 1);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function invalidateAll() {
    foreach ($this->getAll() as $cid) {
      $this->bucket->touch($this->getBinKey($cid), REQUEST_TIME - 1);
    }
  }

}
